; ***********************************************************************
; ***********************************************************************
; *                                                                     *
; *                     Screen Generator Program                        *
; *                                                                     *
; ***********************************************************************
; ***********************************************************************

.386p 
.model flat 
.code 
include Gb.Inc

VStart equ      (16+256*16)                     ; position of first pixel

; ***********************************************************************
; *                                                                     *
; *             Render the current scan line to the screen              *
; *                                                                     *
; *    this draws the background ONLY. The window is drawn separately   *
; *  this renders the screen a line at a time into a buffer, which is   *
; *  copied to the physical screen later                                *
; *                                                                     *
; ***********************************************************************

RenderLine      Proc
        push    eax ebx ecx edx ebp edi esi
; =========================================================================
;      if this is line 143, then copy the buffer to the display
; =========================================================================
        cmp     [byte ptr Memory+0FF44h],143    ; check scan line.
        jne     @@NoCopy
        mov     ebp,offset DisplayImage+VStart  ; draw the sprites
        cmp     [FastSprites],0                 ; if in fast sprite mode
        jz      NoFastSprites
        call    DrawSprites
NoFastSprites:
        call    [PaintVector]                   ; copy logical->physical
@@NoCopy:
; =========================================================================
;       if this is line 0 then generate the scan sprite table
; =========================================================================
        cmp     [byte ptr Memory+0FF44h],0      ; scanline 0 ?
        jne     @@NoSSTable
        cmp     [FastSprites],0                 ; not if fast sprite mode.
        jnz     @@NoSSTable
        call    SSBuildTable                    ; get & sort for this frame
@@NoSSTable:
; =========================================================================
;     if we are using scan line sprites,extract sprites for this line
; =========================================================================
        cmp     [FastSprites],0                 ; not if fast sprite mode.
        jnz     @@NoFastSpr2
        call    SSExtractLine                   ; get em out.
@@NoFastSpr2:
; =========================================================================
; calculate the position in the display buffer - this is scanline * 256+16
; the shift is so we can scroll it back
; =========================================================================
        mov     ch,[byte ptr Memory+0FF44h]     ; cx = scan line # * 256
        mov     ebp,ecx
        and     ebp,00000FF00h
        add     ebp,offset DisplayImage+VStart  ; start rendering 16 in
        mov     al,[byte ptr Memory+0FF43h]     ; fine x scrolling
        and     eax,7
        sub     ebp,eax
; =========================================================================
; make EDI point to the first tile number in memory as an offset.
; The line in the virtual background is (scan#+scrolly) & 255. We divide
; this by 8 to give a virtual tile, then multiply by 32 to give a tile offset
; then its converted to an offset to a real physical address. This is
; because the lower 5 bits of this address are used to wrap around the
; background in the horizontal plane.
; =========================================================================
        mov     ch,[byte ptr Memory+0FF44h]     ; ch = scan line
        add     ch,[byte ptr Memory+0FF42h]     ; add vertical scroll
        movzx   edi,ch                          ; edi=(scan line #+scroll)/8
        shr     edi,3
        shl     edi,5                           ; x 32 + 9800h
        add     edi,9800h
        test    [byte ptr Memory+0FF40h],008h   ; if bit 3 of LCDCONT set
        jz      LCDCBit3Skip
        add     edi,400h                        ; then point to mem+9C00h
LCDCBit3Skip:
        mov     al,[byte ptr Memory+0FF43h]     ; ebx = coarse scroll
        movzx   ebx,al                          ; does tile scrolls
        shr     ebx,3                           ; (every 8 pixels)
        add     edi,ebx
; =========================================================================
; ESI is the tile table base. This is either 8000 or 8800 plus the
; remainder of the (virtual background line/8)*2, which puts it in the
; correct position in the tile. (each tile has 16 bits and each 2
; consecutive bits makes up the tile)
; =========================================================================
        and     ch,7                            ; get remainder
        movzx   esi,ch                          ; esi = remainder * 2
        add     esi,esi
        add     esi,offset Memory+8800h
        test    [byte ptr Memory+0FF40h],010h   ; if bit 4 of LCDCONT set
        jz      LCDBit4Skip
        sub     esi,800h
LCDBit4Skip:
; =========================================================================
;       write 168 bytes of palette zero to the current line @EBP
; =========================================================================
        push    ebp
        movzx   ebx,[Byte Ptr Memory+0FF47h]    ; expand the background
        and     bl,3                            ; dot data 0 & 1
        mov     eax,[CopyData+ebx*4]            ; get the byte pattern.
        mov     cl,176/16                       ; bytes to write
ClearBuffer:
        mov     [ebp],eax                       ; write it out.
        mov     [ebp+4],eax
        mov     [ebp+8],eax
        mov     [ebp+12],eax
        add     ebp,16
        dec     cl
        jnz     ClearBuffer
        pop     ebp                             ; restore buffer pointer
        test    [byte ptr Memory+0FF40h],080h   ; if LCD not turned on
        jz      NoScreen                        ; don't render anything
; =========================================================================
;                Draw the behind the background sprites
; =========================================================================
        mov     al,128
        call    SSDraw
; =========================================================================
;                               No background ?
; =========================================================================
        test    [byte ptr Memory+0FF40h],1      ; background drawn ?
        jz      NoBackground
; =========================================================================
;     load AX,CX and DX with colours corresponding to colour 1,2 and 3
; =========================================================================
        mov     bl,[byte ptr Memory+0FF47h]     ; load the background palette
        call    LoadPaletteData
; =========================================================================
;          do 21 characters (the additional is for scrolling !)
;        at this point EBP=^Screen ESI=OFFSET to chars,EDI=^Tiles
; =========================================================================
        mov     [RenderCounter],21              ; 21 characters to render
        test    [byte ptr Memory+0FF40h],20h    ; window on ?
        jz      RenderLoop                      ; if not do the whole line
        mov     bl,[byte ptr Memory+0FF44h]     ; bl = scan line #
        cmp     bl,[byte ptr Memory+0FF4Ah]     ; if < window y
        jb      RenderLoop                      ; (unsigned comparison)
        mov     bl,[byte ptr Memory+0FF4bh]     ; bl = window x
        shr     bl,3                            ; we don't have to draw
        cmp     bl,0                            ; the whole window.
        jg      NotLE1
        mov     bl,0
NotLE1:
        add     bl,2
        mov     [RenderCounter],bl
RenderLoop:
        xor     ebx,ebx                         ; ebx = tile number
        mov     bl,[Byte Ptr Memory+edi]        ; edi is an offset...
        test    [byte ptr Memory+0FF40h],010h   ; if LCDCont bit 4 is clear
        jnz     NoAdjustment                    ; (tiles from 8800)
        add     bl,080h                         ; tile number is signed.
NoAdjustment:
        shl     ebx,4                           ; multiply by 16
        add     ebx,esi                         ; ebx points to the actual tile
        mov     bx,[ebx]                        ; read in the pattern data.
        call    RenderByte                      ; render 8 bits in bx to EBP
        inc     edi                             ; increment the tile pointer
        test    edi,01Fh                        ; wrap around 32 bytes
        jnz     NoWrapAround                    ; this is for scrolling off
        sub     edi,020h                        ; right the inc edi just
NoWrapAround:                                   ; changes bits 4..0.
        dec     [RenderCounter]                 ; do this 21 times
        jnz     RenderLoop
NoBackground:
; =========================================================================
;       Draw the on top sprites if on top bgr & behind window
; =========================================================================
        mov     al,[byte ptr Memory+0FF40h]     ; e.g. if the priority
        test    al,2                            ; bit is zero (window
        jnz     NotBetween                      ; overlaps all) do it
        mov     al,0                            ; now
        call    SSDraw
NotBetween:
; =========================================================================
; is this a window line ? it must be enabled and the y position must be
; greater than or equal to the scanline.
; =========================================================================
        mov     al,[byte ptr Memory+0FF44h]     ; al = scan line #
        cmp     al,[byte ptr Memory+0FF4Ah]     ; if < window y

        jb      NoWindowLine                    ; (unsigned comparison)
        test    [byte ptr Memory+0FF40h],20h    ; window on ?
        jz      NoWindowLine
        mov     ch,[byte ptr Memory+0FF44h]     ; cx = scan line # * 256
        mov     ebp,ecx
        and     ebp,00000FF00h
        add     ebp,offset DisplayImage+VStart  ; start rendering 16 in
; =========================================================================
;             make EBP point to the FIRST pixel in the window
; =========================================================================
        movzx   eax,[byte ptr Memory+0FF4Bh]    ; work out where it starts.
        cmp     al,167                          ; check range
        ja      NoWindowLine
        sub     eax,7                           ; window x
        add     ebp,eax
; =========================================================================
;  check to see if the window is transparent,if not fill with colour zero
;  note : I don't think it's ever transparent, that might be a fault in
;  Marat's FAQ.
; =========================================================================
        push    ebp
        mov     cl,170                          ; dwords to copy = (170-wH)/4
        sub     cl,[byte ptr Memory+0FF4Bh]
        jnc     NoCarry1
        mov     cl,4
NoCarry1:
        shr     cl,2
        inc     cl
        movzx   ebx,[Byte Ptr Memory+0FF47h]    ; expand the background
        and     bl,3                            ; dot data 0 & 1
        mov     eax,[CopyData+ebx*4]            ; get the byte pattern.

        test    [byte ptr Memory+0FF40h],16
        jz      ClearWindow

        add     eax,04040404h                   ; adjust for window bgr
ClearWindow:
        mov     [ebp],eax
        add     ebp,4
        dec     cl
        jnz     ClearWindow
        pop     ebp
WindowNotSolid:
; =========================================================================
;   make EDI point to the tiles in the tile table as (WINDOWY-SCANY)/8*32
; =========================================================================
        mov     al,[Byte Ptr Memory+0FF44h]
        sub     al,[Byte Ptr Memory+0FF4Ah]     ; al = scany-windowy / 8
        mov     ch,al                           ; save the y offset
        movzx   edi,al                          ; in EDI
        shr     edi,3                           ; the tile row
        shl     edi,5                           ; offset into the tile row
        add     edi,offset Memory+9800h         ; point to the table
        test    [byte ptr Memory+0FF40h],040h   ; if bit 6 of LCDCONT set
        jz      LCDCBit6Skip
        add     edi,400h                        ; then point to mem+9C00h
LCDCBit6Skip:
; =========================================================================
;   make ESI be an pointer into the tile table base. This also includes an
;   adjustment factor for the fine y position in the tile
; =========================================================================
        and     ch,7                            ; get remainder
        movzx   esi,ch                          ; esi = remainder * 2
        add     esi,esi
        add     esi,offset Memory+8000h         ; point to table
        test    [byte ptr Memory+0FF40h],16
        jnz     NotWAlt2
        add     esi,800h
NotWAlt2:
; =========================================================================
;     load AX,CX and DX with colours corresponding to colour 1,2 and 3
; =========================================================================
        mov     bl,[byte ptr Memory+0FF47h]     ; load the background palette
        call    LoadPaletteData
        add     eax,04040404h                   ; adjust for window palette
        add     ecx,04040404h                   ; which is colours 14..17
        add     edx,04040404h
NotWalt3:
; =========================================================================
;        at this point EBP=^Screen ESI=^to chars,EDI=^Tiles
; =========================================================================
        push    ebp
        mov     bl,170                          ; chars to copy = (170-wH)/8
        sub     bl,[byte ptr Memory+0FF4Bh]
        jnc     NoCarry2
        mov     bl,8
NoCarry2:
        shr     bl,3
        mov     [RenderCounter],bl              ; 21 characters to render
WindowRenderLoop:
        xor     ebx,ebx                         ; ebx = tile number
        mov     bl,[edi]                        ; read a tile
        test    [byte ptr Memory+0FF40h],16     ; using alternate charset ?
        jnz     NotWAlt
        add     bl,80h                          ; tile number is signed.
NotWAlt:shl     ebx,4                           ; multiply by 16
        add     ebx,esi                         ; ebx points to the actual tile
        mov     bx,[ebx]                        ; read in the pattern data.
        call    RenderByte                      ; render 8 bits in bx to EBP
        inc     edi                             ; increment the tile pointer
        dec     [RenderCounter]                 ; do this 21 times
        jnz     WindowRenderLoop
        pop     ebp

NoScreen:
NoWindowLine:
RLExit: mov     al,[byte ptr Memory+0FF40h]     ; if on top sprites are
        test    al,2                            ; drawn on top of the window
        jz      RLExit2
        test    [byte ptr Memory+0FF40h],080h   ; if LCD not turned on
        jz      RLExit2                         ; don't draw it
        mov     al,0                            ; draw the on top sprites
        call    SSDraw
RLExit2:
        pop     esi edi ebp edx ecx ebx eax
        ret

.data
RenderCounter db 0                              ; render loop counter
.code
EndP RenderLine

; ***********************************************************************
; *                                                                     *
; *     Convert the palette byte in BL to writeable double words in     *
; *     EAX (colour 1) ECX (colour 2) EDX (colour 3)                    *
; *                                                                     *
; ***********************************************************************

LoadPaletteData proc
        push    esi
        shr     bl,2                            ; don't load palette 0
        mov     esi,ebx
        and     esi,3
        mov     eax,[CopyData+esi*4]            ; load colour 1
        shr     bl,2
        mov     esi,ebx
        and     esi,3
        mov     ecx,[CopyData+esi*4]            ; load colour 2
        shr     bl,2
        mov     esi,ebx
        and     esi,3
        mov     edx,[CopyData+esi*4]            ; load colour 3
        pop     esi
        ret
EndP LoadPaletteData

.data
CopyData dd     13131313h,12121212h             ; byte patterns
         dd     11111111h,10101010h
.code

; ***********************************************************************
; *     Render the word in bx to the screen at ebp. bx contains the     *
; *     bits from the character table in 'raw' format - e.g. bl is      *
; *     the low bits of the colours and bh is the high bits of the      *
; *     colours (represents 8 pixels)                                   *
; *     On exit, ebp is advanced by 8 to point to the next pixel to     *
; *     be written on the screen.                                       *
; *     This must preserve ESI EDI EAX ECX EDX                          *
; ***********************************************************************

RenderByte Proc
        cmp     bx,0
        je      NoEffect
        push    esi                     ; use ESI as work.

                                        ; quick translate from NESA !
                                        ; here bh = h7-h4 h3-h0 bl = l7-l4 l3-l0
        ror     bl,4                    ; now  bh = h7-h4 h3-h0 bl = l3-l0 l7-l4
        ror     bx,4                    ; now  bh = l7-l4 h7-h4 bl = h3-h0 l3-l0
        ror     bh,4                    ; now  bh = h7-h4 l7-l4 bl = h3-h0 l3-l0

        movzx   esi,bh                  ; esi = patterns bits 7..4
        call    [Dword Ptr RenderVector+esi*4]
        add     ebp,4                   ; next spot.
        movzx   esi,bl                  ; esi = patterns bits 3..0
        call    [Dword Ptr RenderVector+esi*4]
        add     ebp,4                   ; next spot.
        pop     esi                     ; restore old ESI
        ret
NoEffect:
        add     ebp,8
        ret
EndP RenderByte

include RenGen.Asm                      ; include generated code.

; ***********************************************************************
; *                                                                     *
; *        Copy the physical screen to the logical screen               *
; *                                                                     *
; ***********************************************************************

CopyLogicalPhysical Proc
        mov     edi,[VideoRAM]
        add     edi,84+16*320
        mov     esi,offset DisplayImage+VStart
        mov     dl,144                  ; number of lines to copy
@@Copy1:mov     dh,160/16               ; bytes to copy

@@Copy2:mov     eax,[esi]               ; copy one line
        mov     ebx,4[esi]              ; in 16 byte chunks
        mov     ecx,8[esi]
        mov     ebp,12[esi]
        mov     [edi],eax
        mov     4[edi],ebx
        mov     8[edi],ecx
        mov     12[edi],ebp
        add     esi,16
        add     edi,16
        dec     dh
        jnz     @@Copy2
        add     edi,160
        add     esi,256-160
        dec     dl
        jnz     @@Copy1
        ret
EndP CopyLogicalPhysical

; ***********************************************************************
; *                                                                     *
; *          Copy logical screen to physical, tweak mode                *
; *                                                                     *
; ***********************************************************************

CopyLogicalTweak Proc
        mov     edi,[VideoRAM]
        add     edi,20+4*360
        mov     esi,offset DisplayImage+VStart
        mov     dl,144                  ; number of lines to copy
@@Topy1:mov     dh,160                  ; bytes to copy
@@Topy2:mov     al,[esi]
        mov     ah,al
        mov     [edi],ax
        inc     esi
        add     edi,2
        dec     dh
        jnz     @@Topy2
        add     esi,256-160
        add     edi,40
        dec     dl
        jnz     @@Topy1
        ret
EndP CopyLogicalTweak

; ***********************************************************************
; *                                                                     *
; *                        Expanded copy                                *
; *                                                                     *
; ***********************************************************************

CopyLogicalLarge Proc
        mov     edi,[VideoRam]
        mov     esi,offset DisplayImage+VStart
        mov     ch,3
        mov     dh,144
@@CLL1: mov     dl,160
        push    esi
        push    edi
@@CLL2: mov     al,[esi]
        mov     ah,al
        mov     [edi],ax
        inc     esi
        inc     edi
        inc     edi
        dec     dl
        jnz     @@CLL2
        pop     edi
        pop     esi
        add     edi,320
        add     esi,256
        dec     ch
        jnz     @@CLL3
        mov     ch,80
@@CLL4: mov     eax,[edi-320]
        mov     [edi],eax
        add     edi,4
        dec     ch
        jnz     @@CLL4
        mov     ch,3
@@CLL3: dec     dh
        jnz     @@CLL1
        ret
EndP CopyLogicalLarge
end
