/*

				Translate .CFG files into .C97 files

*/

#include <stdio.h>
#include <string.h>
#include <dir.h>

unsigned long GetColour(char *c);

unsigned long Palette[16];				/* The translated data */
int nLine;     							/* Line read in */

#define ERR(s) fprintf(stderr,"%s at %d.\n",s,nLine)

void Process(char *s)
{
int n;
unsigned long P;
strlwr(s);								/* make it lower case */
if (*s != '-') { ERR("Missing -");return; }
s++;
n = -1;
if (strncmp(s,"bc",2) == 0) n = 0;  	/* where does it come from ? */
if (strncmp(s,"wc",2) == 0) n = 4;
if (strncmp(s,"sc",2) == 0) n = 8;
if (n < 0) { ERR("Unknown palette type"); return;}

s = s+2;                 				/* Get the subsection */
if (*s < '0' || *s > '4')
	{ ERR("Unknown palette part"); return; }
n = n + 3 - ((*s++) - '0');				/* Offset in palette */

while (*s <= ' ' && *s != '\0') s++;	/* skip over spaces */
P = -1L;
if (*s != '#')
	{
	while (*s <= ' ' && *s != '\0') s++;
	P = GetColour(s);
	}
else
	{
	s++;
	if (sscanf(s,"%lx",&P) != 1)			/* get the colour code */
		{ ERR("Unknown colour constant");return; }
	}

Palette[n] = P;              			/* set the palette */
if (n >= 8) Palette[n+4] = P;			/* NB: VGB only has one for sprites */
}

void Translate(char *FileName)
{
FILE *fIn,*fOut;
int i;
long P;
char s[512];
nLine = 0;

for (i = 0;i < 16;i++)					/* Default values */
	{
	P = 0x50505050L * (i & 3);
	Palette[i] = P;
	}

fIn = fopen(FileName,"r");				/* Open filename for input */
if (fIn == NULL)
	{
	fprintf(stderr,"Couldn't open '%s' for translation.\n",FileName);
	return;
	}
printf("Now translating \"%s\".\n",FileName);
while (fgets(s,sizeof(s),fIn) != NULL)
	{
	nLine++;
	while (*s != '\0' && *s <= ' ') strcpy(s,s+1);
	while (*s != '\0' && *(s+strlen(s)-1) <= ' ') *(s+strlen(s)-1) = '\0';
	if (*s != '\0' && *s != ';') Process(s);
	}
fclose(fIn);							/* we're done looking */

strcpy(s,FileName);						/* Get the output file */
strcpy(strchr(s,'.'),".C97");			/* Filename to write */
fOut = fopen(s,"wb");					/* Write it */
if (fOut == NULL)
	{
	fprintf(stderr,"Can't write file \"%s\"\n",s);
	return;
	}
fwrite(&Palette,4,16,fOut);
fclose(fOut);
}

void main(int argc,char *argv[])
{
int i;
struct ffblk ff;
for (i = 1;i < argc;i++)
			Translate(argv[i]);
if (argc == 1)
	{
	i = findfirst("*.cfg",&ff,0);
	while (i == 0)
		{
		Translate(ff.ff_name);
		i = findnext(&ff);
		}
	}
}

struct ColorName_struct
{
 char *name;
 unsigned long color;
};
typedef struct ColorName_struct ColorName_t;

static ColorName_t ColorName[]=
{
 { "black"   ,0x000000L },
 { "blue"    ,0x0000FFL },
 { "green"   ,0x00FF00L },
 { "cyan"    ,0x55FFFFL },
 { "red"     ,0xFF0000L },
 { "magenta" ,0xFF55FFL },
 { "brown"   ,0xAA5500L },
 { "orange"  ,0xFF7711L },
 { "pink"    ,0xFFAAAAL },
 { "gray"    ,0xAAAAAAL },
 { "yellow"  ,0xFFFF00L },
 { "white"   ,0xFFFFFFL },
 { "dkgray"    ,0x555555L },
 { "dkblue"    ,0x0000AAL },
 { "dkgreen"   ,0x00AA00L },
 { "dkcyan"    ,0x00AAAAL },
 { "dkred"     ,0xAA0000L },
 { "dkmagenta" ,0xAA00AAL },
 { NULL,-1 }
};

unsigned long GetColour(char *c)
{
int i = 0;
while (ColorName[i].name != NULL &&
		stricmp(ColorName[i].name,c) != 0) i++;
return(ColorName[i].color);
}